#define UNICODE
#define _UNICODE

#include "windows.h"
#include "perfc.h"

//-----------------------------------------------------------------
BOOL GetCounterNumbers(LPCTSTR pszObjName, LPTSTR pszObjNumber,
                       LPCTSTR pszCntName, LPTSTR pszCntNumber)
{
    TCHAR sz[MAX_PATH], szDir[MAX_PATH], szFile[MAX_PATH];
    LPBYTE buf = NULL, buf1 = NULL;
    HKEY hKey = NULL;
    DWORD rc, dwSize, dwRecv, dwFileSize;
    LPTSTR p, p2;
    HANDLE hFile;
    BOOL status;

    // allocate buffer for performance counter data

    dwSize = CNT_LIST_SIZE;
    buf = malloc(dwSize);
    if (buf == NULL) {
        return FALSE;
    }

    // trying getting the list of counters (using neutral English 
    // version) from the registry, first

    wsprintf(sz, TEXT("%s\\%03x"), REGKEY_PERFC,
             MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

    rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE,sz,0,KEY_READ,&hKey);
    if (rc == ERROR_SUCCESS) {
        memset(buf, 0, dwSize);
        rc = RegQueryValueEx(hKey, REGVAL_COUNTERS, NULL, 
                             NULL, buf, &dwSize);
        if (rc == ERROR_MORE_DATA) {
            buf1 = (LPBYTE)realloc(buf, dwSize);
            if (buf1 == NULL) {
                RegCloseKey(hKey);
                free(buf);
                return FALSE;
            }
            buf = buf1;
            memset(buf, 0, dwSize);
            rc = RegQueryValueEx(hKey, REGVAL_COUNTERS, NULL,
                                 NULL, buf, &dwSize);
        }
        RegCloseKey(hKey);
    }

    if (rc != ERROR_SUCCESS) {
    
        // if not in the registry, try getting the list of counters
        // (using neutral English version) from the perf file.

        GetWindowsDirectory(szDir, MAX_PATH);
        wsprintf(szFile, TEXT("%s\\system32\\perfc%03x.dat"),szDir, 
                 MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

        hFile = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, 
                           NULL, OPEN_EXISTING, 
                           FILE_ATTRIBUTE_NORMAL, NULL);
        if (hFile == NULL) {
            free(buf);
            return FALSE;
        }

        dwFileSize = GetFileSize(hFile, NULL);
        if (dwFileSize > dwSize) {
            dwSize = dwFileSize;
            buf1 = (LPBYTE)realloc(buf, dwSize);
            if (buf1 == NULL) {
                free(buf);
                CloseHandle(hFile);
                return FALSE;
            }
            buf = buf1;
        }

        memset(buf, 0, dwSize);
        if (!ReadFile(hFile, (PVOID)buf, dwSize, &dwRecv, NULL)) {
            free(buf);
            CloseHandle(hFile);
            return FALSE;
        }
        CloseHandle(hFile);
    }

    // search multi_sz buffer for object number/name pair
    
    status = FALSE;
    p = (LPTSTR)buf;
    while (*p) {
        if (p > (LPTSTR)buf) {
            for (p2=p-2; isdigit(*p2); p2--);
        }
        if (lstrcmpi(p, pszObjName) == 0) {
            // look backwards for the counter number
            for (p2=p-2; isdigit(*p2); p2--);
            lstrcpy(pszObjNumber, p2+1);
            status = TRUE;
            break;
        }
        p += (lstrlen(p) + 1);
    }
    if (!status) {
        return FALSE;
    }

    // search multi_sz buffer for counter number/name pair

    status = FALSE;
    p = (LPTSTR)buf;
    while (*p) {
        if (p > (LPTSTR)buf) {
            for (p2=p-2; isdigit(*p2); p2--);
        }
        if (lstrcmpi(p, pszCntName) == 0) {
            // look backwards for the counter number
            for (p2=p-2; isdigit(*p2); p2--);
            lstrcpy(pszCntNumber, p2+1);
            status = TRUE;
            break;
        }
        p += (lstrlen(p) + 1);
    }

    free(buf);
    return status;

} // GetCounterNumbers

//-----------------------------------------------------------------
BOOL GetCounterData(LPTSTR pszObjectNumber, LPTSTR pszInstance, 
                    DWORD dwCounterNumber, LPBYTE lpData, 
                    DWORD dwDataSize)
{
    DWORD rc, dwSize, i,j,k, dwActualSize, dwObjectNumber;
    LPBYTE buf = NULL, buf1 = NULL;
    LPTSTR psz = NULL;
    BOOL status = FALSE;
    PPERF_DATA_BLOCK pPerf;
    PPERF_OBJECT_TYPE pObj;
    PPERF_INSTANCE_DEFINITION pInst;
    PPERF_COUNTER_BLOCK pCntBlock;
    PPERF_COUNTER_DEFINITION pDef, pCurDef;

    // allocate buffer for performance data

    dwSize = CNT_DATA_SIZE;
    buf = malloc(dwSize);
    if (buf == NULL) {
        return FALSE;
    }
    memset(buf, 0, dwSize);

    // query the performance data

    while (TRUE) {
        dwActualSize = dwSize;
        rc = RegQueryValueEx(HKEY_PERFORMANCE_DATA,pszObjectNumber,
                             NULL, NULL, buf, &dwActualSize);
        if (rc == ERROR_SUCCESS) {
            break;
        }
        if (rc == ERROR_MORE_DATA) {
            dwSize += CNT_DATA_SIZE;
            buf1 = realloc(buf, dwSize);
            if (buf1 == NULL) {
                free(buf);
                return FALSE;
            }
            buf = buf1;
            memset(buf, 0, dwSize);
        } else {    // return FALSE for any other error
            free(buf);
            return FALSE;
        }
    }

    // check for valid perf data block signature

    pPerf = (PPERF_DATA_BLOCK)buf;
    if ((dwActualSize < 4) ||
        (pPerf)->Signature[0] != (WCHAR)'P' ||
        (pPerf)->Signature[1] != (WCHAR)'E' ||
        (pPerf)->Signature[2] != (WCHAR)'R' ||
        (pPerf)->Signature[3] != (WCHAR)'F') {

        free(buf);
        RegCloseKey(HKEY_PERFORMANCE_DATA);
        return FALSE;
    }

    // search perf instance data

    dwObjectNumber = _wtoi(pszObjectNumber);
    pObj  = (PPERF_OBJECT_TYPE)((DWORD)pPerf+pPerf->HeaderLength);

    // walk list of objects
    for (i=0; i < pPerf->NumObjectTypes; i++) {

        if (pObj->ObjectNameTitleIndex == dwObjectNumber) {
            // found specified object
            pDef = (PPERF_COUNTER_DEFINITION)
                   ((PBYTE)pObj + pObj->HeaderLength);

            if (pObj->NumInstances > 0) {
                // this object has instances
                pInst = (PPERF_INSTANCE_DEFINITION)
                         ((PBYTE)pObj + pObj->DefinitionLength);

                // walk list of instances
                for (j = 0; j < (DWORD)pObj->NumInstances; j++) {
                    pCurDef = pDef;
                    pCntBlock = (PPERF_COUNTER_BLOCK)
                                ((PBYTE)pInst + pInst->ByteLength);
                    psz = (LPTSTR)((PBYTE)pInst+pInst->NameOffset);

                    if (lstrcmpi(psz, pszInstance) == 0) {
                        // found specified instance,
                        // walk list of counters
                        for (k = 0; k < pObj->NumCounters; k++) {
                            if (pCurDef->CounterNameTitleIndex == 
                                dwCounterNumber) {
                                // found specified counter
                                // copy data to user's buffer
                                if (pDef->CounterSize 
                                    <= dwDataSize) {

                                    memcpy(lpData, 
                                           ((PBYTE)pCntBlock + 
                                           pCurDef->CounterOffset),
                                           pCurDef->CounterSize);
                                    status = TRUE;
                                    goto exit;
                                } else {
                                    status = FALSE;
                                    goto exit;
                                }
                            }
                            pCurDef = (PPERF_COUNTER_DEFINITION)
                                      ((PBYTE)pCurDef + 
                                      pCurDef->ByteLength);
                        }
                    }
                    pInst = (PPERF_INSTANCE_DEFINITION)
                            ((PBYTE)pCntBlock + 
                            pCntBlock->ByteLength);
                }
            } else {
                // this object doesn't have instances
                pCntBlock = (PPERF_COUNTER_BLOCK) 
                            ((PBYTE)pObj + pObj->DefinitionLength);
                
                // walk list of counters
                for (k = 0; k < pObj->NumCounters; k++) {
                    if (pDef->CounterNameTitleIndex == 
                        dwCounterNumber) {
                        // found specified counter
                        // copy data to user's buffer
                        if (pDef->CounterSize <= dwDataSize) {
                            memcpy(lpData, 
                                   ((PBYTE)pCntBlock + 
                                   pDef->CounterOffset),
                                   pDef->CounterSize);
                            status = TRUE;
                            goto exit;
                        } else {
                            status = FALSE;
                            goto exit;
                        }
                    }
                    pDef = (PPERF_COUNTER_DEFINITION)
                           ((PBYTE)pDef + pDef->ByteLength);         
                }
            }
        }
        pObj = (PPERF_OBJECT_TYPE)
               ((LPBYTE)pObj + pObj->TotalByteLength);
    }
    status = FALSE;
exit:    
    free(buf);
    RegCloseKey(HKEY_PERFORMANCE_DATA);
    return TRUE;
} // GetCounterData
