#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <wininet.h>
#include "inetinfo.h"

#ifndef _MT
   #error "No threads in single-threaded app"
#endif

#define POLL_INTERVAL  (     15 * 1000)
#define QUERY_INTERVAL (60 * 60 * 1000)
#define TERM_TIMEOUT   (      5 * 1000)

static HANDLE s_hThread = 0;
static HANDLE s_hEvent  = 0;

BOOL isRunning( void ) {
    return WAIT_TIMEOUT == WaitForSingleObject( s_hEvent, 0 );
}

void enableRefreshBtn( HWND hwnd, BOOL bEnable ) {
    HWND hwndRefresh = GetDlgItem( hwnd, IDC_REFRESH );
    if ( bEnable ) {
        EnableWindow( hwndRefresh, TRUE );
        FORWARD_WM_NEXTDLGCTL( hwnd, 
            hwndRefresh, TRUE, SNDMSG );
    } else {
        FORWARD_WM_NEXTDLGCTL( hwnd,
            GetDlgItem( hwnd, IDC_EXIT ), TRUE, SNDMSG );
        EnableWindow( hwndRefresh, FALSE );
    }
}

BOOL isConnected( void ) {
    DWORD dwFlags = 0;
    return InternetGetConnectedState( &dwFlags, 0 );
}

DWORD WINAPI downloadThread( LPVOID lpParameter ) {
    HWND hwnd = (HWND) lpParameter;
    for ( ;; ) {
        BOOL bOK;
        if ( !isConnected() ) {
            setStatus( hwnd, _T( "Waiting for connection" ) );
            while ( !isConnected() ) {
                const DWORD dwReturn = 
                    WaitForSingleObject( s_hEvent, POLL_INTERVAL );
                if ( WAIT_OBJECT_0 == dwReturn && !isRunning() ) {
                    return 0; //*** FUNCTION EXIT POINT
                }
            }
        }

        enableRefreshBtn( hwnd, FALSE );
        bOK = download( hwnd );
        if ( isRunning() ) {
            enableRefreshBtn( hwnd, TRUE );
        }
        setStatus( hwnd, bOK ? _T( "OK" ) : _T( "Error" ) );
        if ( bOK ) {
            const DWORD dwReturn = 
                WaitForSingleObject( s_hEvent, QUERY_INTERVAL );
            if ( WAIT_OBJECT_0 == dwReturn && !isRunning() ) {
                return 0; //*** FUNCTION EXIT POINT
            }
        }
    }
    // return 0; May need this on some compilers to avoid warning.
}

BOOL startDownload( HWND hwnd ) {
    s_hEvent = CreateEvent( 0, TRUE, FALSE, 0 );
    if ( 0 != s_hEvent ) {
        DWORD dwID = 0; // Required for Win9x, can be 0 for NT
        s_hThread = CreateThread( 
            0, 0, downloadThread, hwnd, 0, &dwID );
        if ( 0 == s_hThread ) {
            CloseHandle( s_hEvent );
            MessageBox( hwnd, 
                _T( "Unable to start download thread." ), 
                APP_NAME, MB_OK | MB_ICONERROR );
        }
    }
    return 0 != s_hThread;
}

void pumpMessages( void ) {
    MSG msg = { 0 };
    while ( PeekMessage( &msg, 0, 0, 0, PM_REMOVE ) ) {
        TranslateMessage( &msg );
        DispatchMessage ( &msg );
    }
}

void waitForThread( void ) {
    const DWORD dwEndOfTime = GetTickCount() + TERM_TIMEOUT;
    while ( GetTickCount() < dwEndOfTime ) {
        const DWORD dwTimeLeft = dwEndOfTime - GetTickCount();
        // WaitForSingleObject would hang the UI!
        const DWORD dwResult = MsgWaitForMultipleObjects(
            1, &s_hThread, FALSE, dwTimeLeft, QS_ALLEVENTS );
        if ( WAIT_TIMEOUT  == dwResult || 
             WAIT_OBJECT_0 == dwResult ||
             WAIT_FAILED   == dwResult ) 
        {
            break; //*** LOOP EXIT POINT
        }
        pumpMessages();
    }
}

void stopDownload( void ) {
    if ( 0 != s_hEvent ) {
        SetEvent( s_hEvent ); // isRunning() now returns false.
        if ( 0 != s_hThread ) {
            waitForThread();
            CloseHandle( s_hThread );
        }
        CloseHandle( s_hEvent );
    }
}

void update( void ) {
    if ( isRunning() ) {
        PulseEvent( s_hEvent );
    }
}

/* end of file */
